#include "compiler.h"
#include "preprocessor.h"
#include "board.h"
#include "gpio.h"
#include "sysclk.h"
#include "serial.h"
#include "conf_usart_serial.h"

#include "adc.h"
#include "nvm.h"
#include "pmic.h"
#include "string.h"

status_code_t uart_puts( USART_t *usart, const char *data );


//! Temperature sensor calibration data.
static uint16_t tempsense;

//! The latest, computed temperature.
static volatile uint16_t last_temperature;

static void adc_handler(ADC_t *adc, uint8_t channel, adc_result_t result)
{
	uint32_t temperature;

	temperature = result; 
//	temperature = (uint32_t)result * 358; 
//	temperature /= tempsense;

	last_temperature = temperature & 0xffff;

	char text[20];
	uart_puts( USART_SERIAL_EXAMPLE, "Temperatur : " );
	itoa( last_temperature, text, 10 );
	uart_puts( USART_SERIAL_EXAMPLE, text );
	uart_puts( USART_SERIAL_EXAMPLE, "\n");

	adc_start_conversion(adc, (1 << channel));
}

status_code_t uart_puts( USART_t *usart, const char *data )
{
	while (*data)
	{
		usart_putchar( usart, *data++ );
	}
	return STATUS_OK;
}

int main(void)
{
	struct adc_config         adc_conf;
	struct adc_channel_config adcch_conf;

	// USART options.
	static usart_serial_options_t USART_SERIAL_OPTIONS =
	{
		 .baudrate     = USART_SERIAL_EXAMPLE_BAUDRATE,
		 .charlength   = USART_SERIAL_CHAR_LENGTH,
		 .paritytype   = USART_SERIAL_PARITY,
		 .stopbits     = USART_SERIAL_STOP_BIT
	};

	board_init();
	sysclk_init();
	pmic_init();
	cpu_irq_enable();

	usart_serial_init(USART_SERIAL_EXAMPLE,&USART_SERIAL_OPTIONS);

	memset(&adc_conf, 0, sizeof(struct adc_config));
	memset(&adcch_conf, 0, sizeof(struct adc_channel_config));

	/* Configure the ADC module:
	 * - unsigned, 12-bit results
	 * - bandgap (1 V) voltage reference
	 * - 200 kHz maximum clock rate
	 * - manual conversion triggering
	 * - temperature sensor enabled
	 * - callback function
	 */
	adc_set_conversion_parameters(&adc_conf, ADC_SIGN_OFF, ADC_RES_12, ADC_REF_BANDGAP);
	adc_set_clock_rate(&adc_conf, 200000UL);
	adc_set_conversion_trigger(&adc_conf, ADC_TRIG_MANUAL, 0, 0);
	adc_enable_internal_input(&adc_conf, ADC_INT_TEMPSENSE);

	adc_write_configuration(&ADCA, &adc_conf);
	adc_set_callback(&ADCA, &adc_handler);

	/* Configure ADC channel 0:
	 * - single-ended measurement from temperature sensor
	 * - interrupt flag set on completed conversion
	 * - interrupts disabled
	 */
	adcch_set_input(&adcch_conf, ADCCH_POS_TEMPSENSE, ADCCH_NEG_NONE, 1);
	adcch_set_interrupt_mode(&adcch_conf, ADCCH_MODE_COMPLETE);
	adcch_enable_interrupt(&adcch_conf);

	adcch_write_configuration(&ADCA, 0, &adcch_conf);

	// Get measurement for 85 degrees C (358 kelvin) from calibration data.
//	tempsense = adc_get_calibration_data(ADC_CAL_TEMPSENSE);

	// Enable the ADC and start the first conversion.
	adc_enable(&ADCA);
	adc_start_conversion(&ADCA, ADC_CH0);

	do
	{
		sleepmgr_enter_sleep();
	}	
	while (true);
}

